import pmm_config
import os
import subproc
import shutil


class PMMUtilityException(subproc.NonzeroExitException):
    def __init__(self, utility, nonzero_exit_exception):
        subproc.NonzeroExitException.__init__(self, nonzero_exit_exception.subprocess,nonzero_exit_exception.exitcode)
        self.__utility = utility

    def utility(self):
        return self.__utility


class PleskActualDump:
    def __init__(self, session_path):
        self.__session_path = session_path
        self.__plesk_actual_dump = None

    def get(self):
        #invoke pleskbackup to get plesk actual dump
        cmd = subproc.CmdLine(pmm_config.backup())
        for plesk_backup_arg in pmm_config.plesk_backup_args():
            cmd.arg(plesk_backup_arg)
        cmd.arg('server')
        cmd.arg('-c')
        cmd.arg('--migration-mode')
        session_path_arg = '--session-path=' + self.__session_path
        cmd.arg(session_path_arg)
        errcode = 0
        try:
            errcode = cmd.spawn()
        except subproc.NonzeroExitException, x:
            raise PMMUtilityException('plesk_agent_manager', x)

        self.__plesk_actual_dump = os.path.join(self.__session_path,"dump.xml")
        return self.__plesk_actual_dump


class ConflictDetector:
    def run(owner_guid, plesk_actual_dump, restore_specification, conflicts_description_out,session_path):
        cmd = subproc.CmdLine(pmm_config.plesk_runner())
        for plesk_runner_arg in pmm_config.plesk_runner_args():
            cmd.arg(plesk_runner_arg)
        cmd.arg(pmm_config.pmm_conflict_detector())
        cmd.arg('--detect-conflicts')
        if not plesk_actual_dump:
            plesk_actual_dump = PleskActualDump(session_path).get()
        plesk_dump_arg = '--plesk-dump=' + plesk_actual_dump
        cmd.arg(plesk_dump_arg)
        if owner_guid:
            owner_guid_arg = '--owner-guid=' + owner_guid
            cmd.arg(owner_guid_arg)
        restore_specification_arg = '--restore-specification=' + restore_specification
        cmd.arg(restore_specification_arg)
        
        conflict_iteration = 0
        conflicts_description_out_based = conflicts_description_out
        while os.path.isfile(conflicts_description_out):
            conflict_iteration += 1
            conflicts_description_out = conflicts_description_out_based + "." + str(conflict_iteration)
        
        if conflicts_description_out_based != conflicts_description_out:
            shutil.copy2(conflicts_description_out_based,conflicts_description_out)
        
        conflicts_description_arg = '--conflicts-description-out=' + conflicts_description_out_based
        cmd.arg(conflicts_description_arg)
        
        session_path_arg = '--session-path=' + session_path
        cmd.arg(session_path_arg)
        try:
            errcode = cmd.spawn()
        except subproc.NonzeroExitException, x:
            raise PMMUtilityException('Conflict Detector', x)

        if not os.path.isfile(conflicts_description_out):
            # suppose there is no conflicts found
            return None
        else:
            return conflicts_description_out

    run = staticmethod(run)


class ConflictResolver:
    def run(owner_guid, plesk_actual_dump, conflicted_restore_specification, restore_specification, conflicts_description,conflicts_resolution_rules,session_path):
        cmd = subproc.CmdLine(pmm_config.plesk_runner())
        for plesk_runner_arg in pmm_config.plesk_runner_args():
            cmd.arg(plesk_runner_arg)
        cmd.arg(pmm_config.pmm_conflict_resolver())
        cmd.arg('--resolve-conflicts')
        if not plesk_actual_dump:
            plesk_actual_dump = PleskActualDump(session_path).get()
        plesk_dump_arg = '--plesk-dump=' + plesk_actual_dump
        cmd.arg(plesk_dump_arg)
        if owner_guid:
            owner_guid_arg = '--owner-guid=' + owner_guid
            cmd.arg(owner_guid_arg)
        restore_specification_arg = '--restore-specification=' + conflicted_restore_specification
        cmd.arg(restore_specification_arg)
        conflicts_description_arg = '--conflicts-description=' + conflicts_description
        cmd.arg(conflicts_description_arg)
        conflicts_resolution_rules_arg = '--conflicts-resolution-rules=' + conflicts_resolution_rules
        cmd.arg(conflicts_resolution_rules_arg)
        new_restore_specification_arg = '--restore-specification-out=' + restore_specification
        cmd.arg(new_restore_specification_arg)
        session_path_arg = '--session-path=' + session_path
        cmd.arg(session_path_arg)
        try:
            errcode = cmd.spawn()
        except subproc.NonzeroExitException, x:
            raise PMMUtilityException('Conflict Resolver', x)

        return restore_specification

    run = staticmethod(run)

