import os
import logging
import osutil
import sys
try: import subprocess
except ImportError: import compat.subprocess as subprocess

import pmm_config


mswindows = (sys.platform == "win32")


_logger = logging.getLogger("pmmcli.daemon")


class PMMCliDaemon:
    def __init__(self):
        pass

    def __get_pmmcli_daemon_pid(self):
        pid = ''
        pid_value = None
        try:
            pidfile = open(os.path.join(pmm_config.pmmcli_dir(),'pmmcli_daemon.pid'))
        except IOError:
            return None
        try:
            pid = pidfile.read( )
        finally:
            pidfile.close()
        try:
            pid_value = int(pid)
        except:
            pass
        return pid_value

    def __start(self):
        _logger.debug("starting pmmcli daemon...")
        if mswindows:
            #
            # Could not use subprocess module here. On Windows system it always creates subprocess with handle inheritance.
            # Use os-native services
            #
            
            from win32process import CreateProcess, STARTUPINFO, NORMAL_PRIORITY_CLASS, STARTF_USESTDHANDLES
            from pywintypes import error
            
            try:
                cmd = pmm_config.daemon()
                for arg in pmm_config.daemon_args():
                    cmd = cmd + osutil.args_separator + arg
                
                startupinfo = STARTUPINFO()
                startupinfo.dwFlags |= STARTF_USESTDHANDLES
                startupinfo.hStdInput = 0
                startupinfo.hStdOutput = 0
                startupinfo.hStdError = 0
                
                hp, ht, pid, tid = CreateProcess(None,
                                                 cmd,
                                                 None, # process security
                                                 None, # thread security
                                                 0, # do not inherit handles
                                                 NORMAL_PRIORITY_CLASS,
                                                 os.environ,
                                                 None, # current directory
                                                 startupinfo)
            except error, e:
                _logger.debug(u'pmmcli daemon error: %s' % repr(e))
            else:
                _logger.debug("pmmcli daemon started")
        else:
            try:
                args = pmm_config.daemon_args()
                args.insert(0,pmm_config.daemon())
                subprocess.Popen(args, close_fds = True, stdin = subprocess.PIPE, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
            except OSError, e:
                _logger.debug(u'pmmcli daemon error: %s' % repr(e))
            else:
                _logger.debug("pmmcli daemon started")

    def start(self):
        _logger.debug("detecting running pmmcli daemon...")
        pid = self.__get_pmmcli_daemon_pid()
        if not pid or not osutil.is_active(pid):
            self.__start()
            return True
        else:
            _logger.debug("pmmcli daemon is running")
        return False

