import time
import random
import shutil
import cStringIO
import os, logging
from xml.dom import minidom
from encoder import EncoderFile
import subproc
import pmm_config
import pmmcli_config
import pmm_dump_overview
import pmm_dump_formatter
import pmm_api_xml_protocols
import pmm_conflict_detector
import pmm_dump_access_service

_logger = logging.getLogger("pmmcli.session")

class RSessionException(Exception):
    def __init__(self, session_id):
        self.__session_id = session_id
        _logger.info("RSessionException generated: " + str(self.__session_id))

    def get_session_id(self):
        return self.__session_id


class PmmcliSession:
    def __init__(self, session_id):
        self.__session_id = session_id
        self.__session_path = os.path.join( pmm_config.restore_session_dir(), self.__session_id)
        # test if session directory exists and raise otherwise
        self.__info_xml_base = os.path.join( self.__session_path, 'base')
        self.__restore_specification = os.path.join( self.__session_path, 'restore.xml')
        self.__owner_guid = os.path.join( self.__session_path, 'owner_guid')
        self.__owner_type = os.path.join( self.__session_path, 'owner_type')
        self.__restore_specification_xml_transform = os.path.join( self.__session_path, 'xml_transform.xml')
        self.__conflicted_restore_specification = os.path.join( self.__session_path, 'restore.xml.conflicted')
        self.__dump_specification = os.path.join( self.__session_path, 'dump_specification')
        self.__dump_overview = os.path.join( self.__session_path, 'dump_overview')
        self.__info_xml = os.path.join( self.__session_path, 'info.xml')
        self.__plesk_actual_dump = os.path.join(self.__session_path,'dump.xml')
        if not os.path.isfile(self.__info_xml):
            raise RSessionException(self.__session_id)
        self.__conflict_description = os.path.join( self.__session_path, 'conflict_description')
        self.__conflict_resolution_rules = os.path.join( self.__session_path, 'conflict_resolution_rules')
        _logger.info("Initialized restore session " + str(self.__session_id))

    def get_session_id(self):
        return self.__session_id

    def get_owner_guid(self):
        # return file content
        owner_guid_value = None
        try:
            f = open(self.__owner_guid, 'rt')
            owner_guid_value = f.read()
            f.close()
        except:
            _logger.error("Could not get owner guid from '" + self.__owner_guid + "'.")
            raise RSessionException(self.__session_id)
        return owner_guid_value

    def get_owner_type(self):
        # return file content
        owner_type_value = None
        try:
            f = open(self.__owner_type, 'rt')
            owner_type_value = f.read()
            f.close()
        except:
            _logger.error("Could not get owner type from '" + self.__owner_type + "'.")
            raise RSessionException(self.__session_id)
        return owner_type_value

    def get_restore_specification_xml_transform(self):
        return self.__restore_specification_xml_transform

    def get_restore_specification(self):
        return self.__restore_specification

    def get_conflicted_restore_specification(self):
        return self.__conflicted_restore_specification

    def get_dump_specification(self):
        # return file name
        return self.__dump_specification

    def get_dump_specification_value(self):
        # return file content
        dump_specification_value = None
        try:
            f = open(self.__dump_specification, 'rt')
            dump_specification_value = f.read()
            f.close()
        except:
            _logger.error("Could not get dump specification from '" + self.__dump_specification +"'.")
            raise RSessionException(self.__session_id)
        return dump_specification_value

    def get_dump_overview(self):
        return self.__dump_overview

    def get_session_path(self):
        return self.__session_path

    def get_info_xml(self):
        return self.__info_xml

    def get_plesk_actual_dump(self):
        return self.__plesk_actual_dump

    def get_info_xml_base(self):
        base = self.__session_path
        try:
            f = open(self.__info_xml_base, 'rt')
            base = f.read()
            f.close()
        except:
            pass
        return base

    def conflicts_resolved(self):
        return os.path.isfile(self.__conflicted_restore_specification)

    def apply_base(self):
        writer = pmm_dump_access_service.XMLSimpleWriter()
        setter = pmm_dump_access_service.XmlElementAttrSetter('migration-dump','base',self.get_info_xml_base())
        xml_processor = pmm_dump_access_service.InfoXMLProcessor( self.get_restore_specification(), pmm_dump_access_service.XmlHandler([setter,writer]) )
        xml_processor.process()
        r_s = open(self.get_restore_specification(),'wt')
        r_s.write(writer.get_content())
        r_s.close()

    def create_log(self, logname):
        filename = os.path.join( self.__session_path, logname)
        log = open(filename,'wt')
        return log

    # create new dir and copy info_xml into it
    def createSession(owner_type, owner_guid, info_xml_base, dump_specification):
        # create new unique session_id
        random.seed()
        session_id = time.strftime('%Y%m%d%H%M%S',time.localtime()) + str(random.randint(0,1000))
        session_path = os.path.join( pmm_config.restore_session_dir(), session_id)
        while os.path.isdir(session_path):
            session_id = time.strftime('%Y%m%d%H%M%S',time.localtime()) + str(random.randint(0,1000))
            session_path = os.path.join( pmm_config.restore_session_dir(), session_id)
        _logger.info("Create restore session " + session_id + " started")
        os.makedirs(session_path)
        shutil.copy2(info_xml_base,os.path.join(session_path,u'info.xml'))
        info_xml_base_file = open(os.path.join(session_path,u'base'), 'wt')
        info_xml_base_file.write(info_xml_base)
        info_xml_base_file.close()
        dump_specification_file = open(os.path.join(session_path,u'dump_specification'), 'wt')
        dump_specification_file.write(dump_specification)
        dump_specification_file.close()
        owner_type_file = open(os.path.join(session_path,u'owner_type'), 'wt')
        owner_type_file.write(owner_type)
        owner_type_file.close()
        owner_guid_file = open(os.path.join(session_path,u'owner_guid'), 'wt')
        owner_guid_file.write(owner_guid)
        owner_guid_file.close()
        # copy info_xml into session_path/info.xml
        session = PmmcliSession(session_id)
        # Next will create well-known restore_specification and dump_overview for session
        pmm_dump_overview.DumpOverviewCreator(session.get_session_id()).create(session.get_info_xml(),session.get_dump_overview())
        # now session has info.xml, R.S and D.O.
        _logger.info("Create restore session " + str(session.get_session_id()) + " done")
        return session

    createSession = staticmethod(createSession)

    # get session directory and initialize session files
    def openSession(session_id):
        return PmmcliSession(session_id)

    openSession = staticmethod(openSession)

    # get name for new session and create directory for it
    def deriveSession(session_id):
        # open old session to get info.xml
        session = PmmcliSession.openSession(session_id)
        _logger.info("Derive restore session from session " + str(session.get_session_id()) + " started")
        derived_session = PmmcliSession.createSession(session.get_owner_type(), session.get_owner_guid(), session.get_info_xml_base(), session.get_dump_specification_value())
        # now session has info.xml, R.S and D.O.
        shutil.copy2(os.path.join(session.get_session_path(),'owner_type'),os.path.join(derived_session.get_session_path(),'owner_type'))
        shutil.copy2(os.path.join(session.get_session_path(),'owner_guid'),os.path.join(derived_session.get_session_path(),'owner_guid'))
        derived_session.set_parent(session_id)
        _logger.info("Derive restore session " + str(derived_session.get_session_id()) + " from session " + str(session.get_session_id()) + " done")
        return derived_session

    deriveSession = staticmethod(deriveSession)

    # used for clone session - copying info.xml, R.S, C.D. to new directory
    # this is for '--resolve-conflicts' call, in this case ConflictResolver changes R.S. and initiate RestoreSpecificationCreator().update()
    def cloneSession(session_id):
        session = PmmcliSession.openSession(session_id)
        _logger.info("Clone restore session from session " + str(session.get_session_id()) + " started")
        cloned_session = PmmcliSession.createSession(session.get_owner_type(), session.get_owner_guid(), session.get_info_xml_base(),session.get_dump_specification_value())
        _logger.info("Clone restore session " + str(cloned_session.get_session_id()) + " from session " + str(session.get_session_id()) + " done")
        
        # copy sesion files (R.S., C.D. and other)
        for session_file in os.listdir(session.get_session_path()):
            if not os.path.isdir(session_file):
                shutil.copy2(os.path.join(session.get_session_path(),session_file),os.path.join(cloned_session.get_session_path(),session_file))
        cloned_session.set_parent(session_id)
        
        return cloned_session

    cloneSession = staticmethod(cloneSession)

    def getDumpOverview(self):
        # read dump_overview file and form class DumpOverview instance
        d_o = pmm_api_xml_protocols.DumpOverview.factory()
        d_o.build(minidom.parse(self.__dump_overview).childNodes[0] )
        return d_o

    def getDumpSpecification(self):
        # read dump_overview file and form class DumpSpecification instance
        dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(self.get_dump_specification_value())
        return dump_specification_formatter.buildXml()
    
    def getConflictDescription(self):
        # read conflict_description file and form class ConflictDescription instance
        c_d = pmm_api_xml_protocols.ConflictsDescription.factory()
        c_d.build(minidom.parse(self.__conflict_description).childNodes[0] )
        return c_d

    def set_parent(self, parent_session_id):
        parent = open(os.path.join(self.get_session_path(),'parent'),'wt')
        parent.write(parent_session_id)
        parent.close()

    def __store_restore_specification(self, restore_specification):
        stored_restore_specification = self.__conflicted_restore_specification
        conflict_iteration = 0
        while os.path.isfile(stored_restore_specification):
            conflict_iteration += 1
            stored_restore_specification = self.__conflicted_restore_specification + "." + str(conflict_iteration)
        shutil.copy2(restore_specification,stored_restore_specification)
        _logger.debug(restore_specification + " copied to " + stored_restore_specification)
        return stored_restore_specification

    def __store_conflict_resolution_rules(self, conflict_resolution_rules):
        stored_conflict_resolution_rules = self.__conflict_resolution_rules
        conflict_iteration = 0
        while os.path.isfile(stored_conflict_resolution_rules):
            conflict_iteration += 1
            stored_conflict_resolution_rules = self.__conflict_resolution_rules + "." + str(conflict_iteration)
        f = open(stored_conflict_resolution_rules, 'wt')
        f.write(conflict_resolution_rules)
        f.close()
        return stored_conflict_resolution_rules

    def __doDetectConflicts(self, plesk_actual_dump):
        owner_guid = None
        if self.get_owner_guid() != '':
            owner_guid = self.get_owner_guid()
        pmm_conflict_detector.ConflictDetector.run(owner_guid,plesk_actual_dump,self.__restore_specification,self.__conflict_description,self.__session_path)
        conflicts_description = self.getConflictDescription()
        return len(conflicts_description.get_conflict())

    def detectConflicts(self, plesk_actual_dump):
        return self.__doDetectConflicts(plesk_actual_dump)

    def __doResolveConflictsOnce(self, plesk_actual_dump, conflict_resolution_rules):
        stored_conflict_resolution_rules = self.__store_conflict_resolution_rules(conflict_resolution_rules)
        stored_restore_specification = self.__store_restore_specification(self.__restore_specification)
        owner_guid = None
        if self.get_owner_guid() != '':
            owner_guid = self.get_owner_guid()
        pmm_conflict_detector.ConflictResolver.run(owner_guid, plesk_actual_dump, stored_restore_specification, self.__restore_specification, self.__conflict_description, stored_conflict_resolution_rules, self.__session_path)

    def resolveConflictsOnce(self, plesk_actual_dump, conflict_resolution_rules):
        self.__doResolveConflictsOnce(plesk_actual_dump, conflict_resolution_rules)
        pmm_dump_overview.RestoreSpecificationCreator(self.get_session_id()).update(self.get_info_xml(), self.__dump_overview)

    def resolveConflicts(self, plesk_actual_dump, conflict_resolution_rules):
        iteration = 0
        total_conflicts_count = 0
        max_conflict_resolves = pmmcli_config.get().get_max_conflict_resolves()
        while iteration < max_conflict_resolves:
            _logger.debug("doResolveConflictsOnce called " + str(iteration+1) + " time")
            self.__doResolveConflictsOnce(plesk_actual_dump, conflict_resolution_rules)
            total_conflicts_count = self.__doDetectConflicts(plesk_actual_dump)
            _logger.debug("doResolveConflictsOnce: " + str(total_conflicts_count) + " conflicts total found")
            if total_conflicts_count == 0:
                break
            iteration += 1
        pmm_dump_overview.RestoreSpecificationCreator(self.get_session_id()).update(self.get_info_xml(), self.__dump_overview)
        return total_conflicts_count

    def checkSign(self):
        # invoke sign checker
        backupSignProc = None
        backupSignCmd = subproc.CmdLine(pmm_config.backup_sign())
        backupSignCmd.arg('verify')
        backupSignCmd.arg(self.get_info_xml())
        restore = open(self.get_restore_specification(), 'wt')
        backup_sign_errcode = 0
        try:
            backupSignProc = backupSignCmd.spawn()
            restore.write(backupSignProc.stdout.encode('utf-8'))
        except subproc.NonzeroExitException, x:
            backup_sign_error_codes = [1,2,3]
            backup_sign_errcode = x.exitcode
            if not backup_sign_errcode in backup_sign_error_codes:
                raise pmm_conflict_detector.PMMUtilityException('backup-sign', x) 
            if x.subprocess.stdout:
                restore.write(x.subprocess.stdout.encode('utf-8'))
        restore.close()
        return backup_sign_errcode

    def create_restore_specification(self, selected_objects, owner_guid, owner_type):
        selected_objects_file = None
        self.__apply_secure_transform()
        if selected_objects:
            f = open(self.get_restore_specification_xml_transform(), 'wt')
            resp_str = cStringIO.StringIO()
            resp_encoded = EncoderFile(resp_str, "utf-8")
            resp_encoded.write('<?xml version="1.0" encoding="UTF-8"?>\n')
            selected_objects.export(resp_encoded, 0, name_ = 'transformation')
            f.write(resp_str.getvalue())
            f.close()
            selected_objects_file = self.get_restore_specification_xml_transform()
        pmm_dump_overview.RestoreSpecificationCreator(self.get_session_id()).create(self.get_owner_type(), self.get_owner_guid(), self.get_restore_specification(),self.get_restore_specification(),self.__dump_overview,selected_objects_file)

    # cut only owner hive from restore.xml for security purporses
    def __apply_secure_transform(self):
        owner_type = self.get_owner_type()
        owner_guid = self.get_owner_guid()
        if owner_type != '' and owner_type != 'server' and owner_guid != '':
            # apply secure transform only if restore dump contains owner itself
            # else suppose it is backup of one of owner's child
            
            element_finder = pmm_dump_access_service.XmlElementAttrValueGetter(owner_type,{"guid":owner_guid})
            xml_processor = pmm_dump_access_service.InfoXMLProcessor( self.get_restore_specification(), pmm_dump_access_service.XmlHandler([element_finder]) )
            xml_processor.process()
            
            if element_finder.get_element():
                _logger.debug("Prepare secure transform: element found")
                secure_transform_file = os.path.join( self.__session_path, 'secure_transform')
                secure_transform = open(secure_transform_file, 'wt')
                xml_transform = '<transformation><node children-processing-type="copy" name="%s"><attributes><attribute name="guid" value="%s"/></attributes></node></transformation>' % (self.get_owner_type(),self.get_owner_guid())
                secure_transform.write('<?xml version="1.0" encoding="UTF-8"?>\n')
                secure_transform.write(xml_transform)
                secure_transform.close()
                writer = pmm_dump_access_service.XMLSimpleWriter()
                xml_processor = pmm_dump_overview.InfoXMLProcessor( self.get_restore_specification(),
                                                                    pmm_dump_overview.InfoXmlHandler( pmm_dump_overview.RestoreSpecificationXMLTransformWorker(pmm_dump_overview.RestoreSpecificationXMLTransformSpecification(secure_transform_file),
                                                                                                                                                               writer),
                                                                                                      pmm_dump_overview.XMLTransformPathProcessor()) )
                xml_processor.process()
                result_restore_specification_file = self.get_restore_specification()
                result_restore_specification = open(result_restore_specification_file,'wt')
                result_restore_specification.write(writer.get_content())
                result_restore_specification.close()
                
            else:
                _logger.debug("Owner element for secure transform does not found")
                secure_transform_file = os.path.join( self.__session_path, 'secure_transform')
                secure_transform = open(secure_transform_file, 'wt')
                #hierarchical_nodes = ['admin','reseller','client','domain']
                hierarchical_cut_nodes = {'admin':[],'reseller':['admin','resellers'],'client':['admin','resellers','clients'],'domain':['admin','resellers','clients','client','domains']}
                hierarchical_copy_nodes = {'admin':['admin'],'reseller':['reseller','clients','client','domains','domain'],'client':['client','domains','domain'],'domain':['domain']}
                
                xml_transform = '<transformation>'
                xml_transform_cut_node = '<node children-processing-type="skip" name="%s"/>'
                xml_transform_copy_node = '<node children-processing-type="copy" name="%s"/>'
                
                index = hierarchical_cut_nodes.keys().index(owner_type)
                if index <= 0:
                    raise Exception("Unknown owner type '%s' for restore" % owner_type)
                
                cutting_nodes_array = hierarchical_cut_nodes[owner_type]
                for cutting_node in cutting_nodes_array:
                    cutting_node_transform = xml_transform_cut_node % cutting_node
                    xml_transform = xml_transform + cutting_node_transform
                
                copying_nodes_array = hierarchical_copy_nodes[owner_type]
                for copying_node in copying_nodes_array:
                    copy_node_transform = xml_transform_copy_node % copying_node
                    xml_transform = xml_transform + copy_node_transform
                
                xml_transform = xml_transform + '</transformation>'
                secure_transform.write('<?xml version="1.0" encoding="UTF-8"?>\n')
                secure_transform.write(xml_transform)
                secure_transform.close()
                writer = pmm_dump_access_service.XMLSimpleWriter()
                xml_processor = pmm_dump_overview.InfoXMLProcessor( self.get_restore_specification(),
                                                                    pmm_dump_overview.InfoXmlHandler( pmm_dump_overview.RestoreSpecificationXMLTransformWorker(pmm_dump_overview.RestoreSpecificationXMLTransformSpecification(secure_transform_file),
                                                                                                                                                               writer),
                                                                                                      pmm_dump_overview.XMLTransformPathProcessor()) )
                xml_processor.process()
                result_restore_specification_file = self.get_restore_specification()
                result_restore_specification = open(result_restore_specification_file,'wt')
                result_restore_specification.write(writer.get_content())
                result_restore_specification.close()


